package cmsc420_f22;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

//---------------------------------------------------------------------
// Author: Dave Mount
// For: CMSC 420
// Date: Fall 2022
//
// This program implements a variant of a kd-tree called an XkdTree, 
// for "extended kd-tree". This is an extended tree, meaning that data 
// is stored in the external (leaf) nodes only and internal nodes store 
// splitting information. There is a global quantity called the bucket 
// size, which indicates the maximum number of points that may be stored 
// in any leaf.
//
// Whenever the number of points in an external node exceeds the bucket 
// size, we partition it as follows. We split the longer side of the
// bounding box of the points through the median along this dimension.
//---------------------------------------------------------------------

public class XkdTree<LPoint extends LabeledPoint2D> {

	// -----------------------------------------------------------------
	// Generic Node Type
	// -----------------------------------------------------------------

	private abstract class Node { // generic node type

		abstract LPoint find(Point2D q); // find point in subtree

		abstract Node bulkInsert(List<LPoint> pts) throws Exception; // insert points

		abstract ArrayList<String> list(); // list entries in reverse preorder

		abstract LPoint nearestNeighbor(Point2D center, Rectangle2D cell, LPoint best); // nearest neighbor 

		abstract public String toString(); // string representation
	}

	// -----------------------------------------------------------------
	// Internal node
	// -----------------------------------------------------------------

	private class InternalNode extends Node {

		int cutDim; // the cutting dimension (0 = x, 1 = y)
		double cutVal; // the cutting value
		Node left, right; // children

		/**
		 * Constructor
		 */
		InternalNode(int cutDim, double cutVal, Node left, Node right) {
			this.cutDim = cutDim;
			this.cutVal = cutVal;
			this.left = left;
			this.right = right;
		}

		boolean onLeft(Point2D pt) { // in the left subtree? (for points)
			return pt.get(cutDim) <= cutVal;
		}

		boolean onRight(LPoint pt) { // in the left subtree? (for Labeled points)
			return pt.get(cutDim) >= cutVal;
		}

		boolean onRight(Point2D pt) { // in the left subtree? (for points)
			return pt.get(cutDim) >= cutVal;
		}

		
		Rectangle2D leftPart(Rectangle2D cell) { // left part of cell
			return cell.leftPart(cutDim, cutVal);
		}

		Rectangle2D rightPart(Rectangle2D cell) { // right part of cell
			return cell.rightPart(cutDim, cutVal);
		}

		/**
		 * Find point in this subtree
		 */
		LPoint find(Point2D q) {
			LPoint result = null;
			if (onLeft(q)) {
				result = left.find(q);
			}
			if (result == null && onRight(q)) {
				result = right.find(q);
			}
			return result;
		}
		
		/**
		 * Insert points into this subtree
		 */
		Node bulkInsert(List<LPoint> pts) throws Exception {
			sortPts(pts, cutDim); // sort the points along cutting dimension
			int i = 0;
			while (i < pts.size() && !onRight(pts.get(i))) i++; // find partition index
			List<LPoint> leftPart = pts.subList(0, i); // partition the list
			List<LPoint> rightPart = pts.subList(i, pts.size());
			left = left.bulkInsert(leftPart); // insert each part
			right = right.bulkInsert(rightPart);
			return this;
		}

		/**
		 * Get a list of the nodes in preorder. Adds current key in parentheses,
		 * followed by left and right subtrees.
		 */
		ArrayList<String> list() {
			ArrayList<String> list = new ArrayList<String>();
			list.add(toString()); // add this node
			list.addAll(right.list()); // add right
			list.addAll(left.list()); // add left
			return list;
		}

		/**
		 * Nearest neighbor. Returns the closest point to center. The best
		 * point seen so far is best. We skip the node if its cell is farther
		 * from the best so far. We search both subtrees in order according
		 * to which side is closer. (The version shown here is a bit simpler
		 * than the one in the lecture notes.)
		 */
		LPoint nearestNeighbor(Point2D center, Rectangle2D cell, LPoint best) {
			double bestDist = (best == null ? Double.POSITIVE_INFINITY : center.distanceSq(best.getPoint2D()));
			if (cell.distanceSq(center) <= bestDist) { // is anything here possibly better?
				if (onLeft(center)) { // we are left of center?
					best = left.nearestNeighbor(center, leftPart(cell), best);
					best = right.nearestNeighbor(center, rightPart(cell), best);
				} else { // we are right of center
					best = right.nearestNeighbor(center, rightPart(cell), best);
					best = left.nearestNeighbor(center, leftPart(cell), best);
				}
			}
			return best;
		}

		/**
		 * String representation.
		 */
		public String toString() {
			String cutIndic = (cutDim == 0 ? "x=" : "y=");
			return "(" + cutIndic + cutVal + ")";
		}
	}

	// -----------------------------------------------------------------
	// External node
	// -----------------------------------------------------------------

	private class ExternalNode extends Node {

		ArrayList<LPoint> points; // points stored in this node

		/**
		 * Default constructor.
		 */
		ExternalNode() {
			points = new ArrayList<LPoint>();
		}

		/**
		 * Find point in external node.
		 */
		LPoint find(Point2D q) {
			LPoint result = null;
			for (LPoint p : points) {
				if (p.getPoint2D().equals(q)) {
					result = p;
				}
			}
			return result;
		}
		
		/**
		 * Bulk insert. Inserts a set of points. It computes the longer
		 * side of the bounding box to determine the cutting dimension. The
		 * cut is then made through the median along this dimension.
		 */
		Node bulkInsert(List<LPoint> pts) throws Exception {
			points.addAll(pts); // add to bucket
			if (points.size() <= bucketSize) { // no need to split
				return this;
			} else {
				Rectangle2D box = new Rectangle2D();
				for (LPoint p : points) { // bounding box for points
					box.expand(p.getPoint2D());
				}
				int cutDim = getWideSide(box); // get cutting dimension
				sortPts(points, cutDim); // sort along cutting dimension
				int medIdx = points.size()/2; // median index
				double cutVal;
				if (points.size() % 2 == 0) { // compute cutting value
					cutVal = (points.get(medIdx-1).get(cutDim) + points.get(medIdx).get(cutDim))/2;
				} else {
					cutVal = points.get(medIdx).get(cutDim);
				}

				InternalNode u = new InternalNode(cutDim, cutVal, null, null); // create internal node
				ExternalNode left = new ExternalNode(); // create empty external nodes
				ExternalNode right = new ExternalNode();
				// insert into these nodes
				u.left = left.bulkInsert(points.subList(0, medIdx));
				u.right = right.bulkInsert(points.subList(medIdx, points.size()));
				return u;
			}
		}

		/**
		 * List the contents of this node
		 */
		ArrayList<String> list() {
			Collections.sort(points, new ByLabel());
			ArrayList<String> list = new ArrayList<String>();
			list.add(toString()); // add this node
			return list;
		}

		/**
		 * Nearest neighbor. Returns the closest point to center. 
		 */
		LPoint nearestNeighbor(Point2D center, Rectangle2D cell, LPoint best) {
			double bestDist = (best == null ? Double.POSITIVE_INFINITY : center.distanceSq(best.getPoint2D()));
			for (LPoint p : points) {
				double pDist = center.distanceSq(p.getPoint2D());
				if (pDist < bestDist) { // new best?
					best = p;
					bestDist = pDist;
				}
			}
			return best;
		}

		/**
		 * String representation.
		 */
		public String toString() {
			String contents = new String();
			contents = "[ ";
			for (LPoint p : points) {
				contents += "{" + p.toString() + "} ";
			}
			contents += "]";
			return contents;
		}

	}

	// -----------------------------------------------------------------
	// Tree private data
	// -----------------------------------------------------------------

	private int bucketSize; // max points per leaf
	private Rectangle2D bbox; // the bounding box
	private Node root; // root of the tree
	private int nPoints; // number of points in the tree

	// -----------------------------------------------------------------
	// Tree public members
	// -----------------------------------------------------------------

	/**
	 * Creates an empty tree.
	 */
	public XkdTree(int bucketSize, Rectangle2D bbox) {
		this.bucketSize = bucketSize;
		this.bbox = new Rectangle2D(bbox);
		clear();
	}

	/**
	 * Clear the tree.
	 */
	public void clear() {
		root = new ExternalNode();
		nPoints = 0;
	}

	/**
	 * Number of points.
	 */
	public int size() {
		return nPoints;
	}

	/**
	 * Find an point in the tree.
	 */
	public LPoint find(Point2D q) {
		return root.find(q);
	}

	/**
	 * Insert a point.
	 */
	public void insert(LPoint pt) throws Exception {
		ArrayList<LPoint> pts = new ArrayList<LPoint>();
		pts.add(pt);
		bulkInsert(pts); // reduce to bulk insert
	}

	/**
	 * Bulk insert a set of points.
	 */
	public void bulkInsert(ArrayList<LPoint> pts) throws Exception {
		for (LPoint p : pts) {
			if (!bbox.contains(p.getPoint2D())) {
				throw new Exception("Attempt to insert a point outside bounding box");
			}
		}
		root = root.bulkInsert(pts); // insert the points
		nPoints += pts.size(); // update point count
	}
	
	/**
	 * Delete a point.
	 */
	public void delete(Point2D pt) throws Exception {
		// Not implemented in this version
	}

	/**
	 * Get a list of entries in preorder
	 */
	public ArrayList<String> list() {
		return root.list();
	}

	/**
	 * Nearest neighbor query.
	 */
	public LPoint nearestNeighbor(Point2D center) {
		LPoint best = null;
		LPoint nn = root.nearestNeighbor(center, bbox, best);
		return nn;
	}
	
	// -----------------------------------------------------------------
	// Tree Utilities for sorting and splitting.
	// -----------------------------------------------------------------

	/**
	 * Get the wider side of a rectangle.
	 */
	private static int getWideSide(Rectangle2D cell) {
		return (cell.getWidth(0) >= cell.getWidth(1) ? 0 : 1);
	}
	
	/**
	 * Sort points according to given dimension.
	 */
	private void sortPts(List<LPoint> pts, int d) {
		if (d == 0) { // sort by the cutting dimension
			Collections.sort(pts, new ByXThenY());
		} else {
			Collections.sort(pts, new ByYThenX());
		}
	}

	/**
	 * Comparators for sorting (by (X,y), by (y,x), or by label).
	 */
	private class ByXThenY implements Comparator<LPoint> { // lexicographic (x,y)
		public int compare(LPoint pt1, LPoint pt2) {
			double x1 = pt1.getX();
			double x2 = pt2.getX();
			if (x1 < x2)
				return -1;
			else if (x1 > x2)
				return +1;
			else {
				double y1 = pt1.getY();
				double y2 = pt2.getY();
				if (y1 < y2)
					return -1;
				else if (y1 > y2)
					return +1;
				else {
					return 0;
				}
			}
		}
	}

	private class ByYThenX implements Comparator<LPoint> { // lexicographic (y,x)
		public int compare(LPoint pt1, LPoint pt2) {
			double y1 = pt1.getY();
			double y2 = pt2.getY();
			if (y1 < y2)
				return -1;
			else if (y1 > y2)
				return +1;
			else {
				double x1 = pt1.getX();
				double x2 = pt2.getX();
				if (x1 < x2)
					return -1;
				else if (x1 > x2)
					return +1;
				else
					return 0;
			}
		}
	}

	private class ByLabel implements Comparator<LPoint> { // based on the label
		public int compare(LPoint pt1, LPoint pt2) {
			String s1 = pt1.getLabel();
			String s2 = pt2.getLabel();
			return s1.compareTo(s2);
		}
	}
}
